@@ -10,7 +10,10 @@ import sys
 try:
     from cStringIO import StringIO
 except ImportError:
-    from StringIO import StringIO
+    try:
+        from StringIO import StringIO
+    except ImportError:
+        from io import StringIO
 import token
 import tokenize
 from optparse import OptionParser
@@ -26,7 +29,7 @@ except ImportError:
     pep8 = None
 
 
-__version__ = '0.4'
+__version__ = '0.5'
 
 
 pep8bin = 'pep8'
@@ -77,7 +80,6 @@ class FixPEP8(object):
         self.results = []
         self.options = options
         self.indent_word = _get_indentword("".join(self.source))
-        self.is_found_e111 = False
         # method definition
         self.fix_e222 = self.fix_e221
         self.fix_e223 = self.fix_e221
@@ -89,13 +91,14 @@ class FixPEP8(object):
             if t[0] == token.INDENT:
                 indent_word = t[1]
                 break
-        indent_level = len(indent_word) / len(self.indent_word)
+        import math
+        indent_level = int(math.ceil(float(len(indent_word)) /
+                                     len(self.indent_word)))
         return indent_level + 1
 
     def _spawn_pep8(self, targetfile):
         """execute pep8 via subprocess.Popen."""
         paths = os.environ['PATH'].split(':')
-        paths.reverse()
         for path in paths:
             if os.path.exists(os.path.join(path, pep8bin)):
                 cmd = ([os.path.join(path, pep8bin)] +
@@ -124,21 +127,26 @@ class FixPEP8(object):
                  if self.options.ignore else []))
 
     def _fix_source(self):
+        completed_lines = []
         for result in self.results:
-            if result['id'] == "E111" and self.is_found_e111:
+            if result['line'] in completed_lines:
                 continue
             fixed_methodname = "fix_%s" % result['id'].lower()
             if hasattr(self, fixed_methodname):
                 fix = getattr(self, fixed_methodname)
-                fix(result)
+                modified_lines = fix(result)
+                if modified_lines:
+                    completed_lines += modified_lines
+                completed_lines.append(result['line'])
             else:
-                print >> sys.stderr, "'%s' is not defined." % fixed_methodname
                 if self.options.verbose:
+                    sys.stderr.write("'%s' is not defined.\n" % \
+                            fixed_methodname)
                     info = result['info'].strip()
-                    print >> sys.stderr, "%s:%s:%s:%s" % (result['filename'],
-                                                          result['line'],
-                                                          result['column'],
-                                                          info)
+                    sys.stderr.write("%s:%s:%s:%s\n" % (result['filename'],
+                                                        result['line'],
+                                                        result['column'],
+                                                        info))
 
     def _fix_whitespace(self, result, pattern, repl):
         target = self.source[result['line'] - 1]
@@ -150,15 +158,7 @@ class FixPEP8(object):
             pep8result = self._execute_pep8(self.filename)
         else:
             pep8result = self._spawn_pep8(self.filename)
-        raw_results = [_analyze_pep8result(line) for line in pep8result]
-
-        # Only handle one error per line
-        line_results = {}
-        for result in raw_results:
-            line_results[result['line']] = result
-
-        self.results = line_results.values()
-
+        self.results = [_analyze_pep8result(line) for line in pep8result]
         self._fix_source()
         return "".join(self.source)
 
@@ -173,7 +173,6 @@ class FixPEP8(object):
     #    self.source[result['line'] - 1] = fixed
 
     def fix_e111(self, result):
-        self.is_found_e111 = True
         sio = StringIO("".join(self.source[result['line'] - 1:]))
         last_line = ""
         diff_cnt = 0
@@ -193,6 +192,10 @@ class FixPEP8(object):
         for offset, fixed_line in enumerate(fixed_lines):
             self.source[result['line'] - 1 + offset] = fixed_line
 
+        # Mark everything as modified since we don't want other instances of
+        # E111 fixes to interfere with this fix.
+        return range(len(self.source))
+
     def fix_e201(self, result):
         self._fix_whitespace(result, r"\(\s+", "(")
         self._fix_whitespace(result, r"\[\s+", "[")
@@ -224,7 +227,6 @@ class FixPEP8(object):
 
     def fix_e224(self, result):
         target = self.source[result['line'] - 1]
-        c = result['column'] + 1
         fixed = re.sub(r'\t+', ' ', target, 1)
         self.source[result['line'] - 1] = fixed
 
@@ -258,7 +260,7 @@ class FixPEP8(object):
         target = self.source[result['line'] - 1]
         c = result['column']
 
-        # pep8 is sometiems off by one in cases like "{# comment"
+        # pep8 is sometimes off by one in cases like "{# comment"
         if target[c] == '#':
             pass
         elif target[c - 1] == '#':
@@ -270,7 +272,14 @@ class FixPEP8(object):
         self.source[result['line'] - 1] = fixed
 
     def fix_e262(self, result):
-        self._fix_whitespace(result, r"##* *", "# ")
+        target = self.source[result['line'] - 1]
+        split = target.rsplit('#', 1)
+
+        assert len(split) == 2
+        comment = split[1].lstrip()
+        fixed = split[0].rstrip(' \t#') + ('  # ' + comment if comment else '')
+
+        self.source[result['line'] - 1] = fixed
 
     def fix_e301(self, result):
         cr = self.newline
@@ -281,6 +290,11 @@ class FixPEP8(object):
         cr = self.newline * add_linenum
         self.source[result['line'] - 1] = cr + self.source[result['line'] - 1]
 
+    def fix_e304(self, result):
+        line = result['line'] - 2
+        if not self.source[line].strip():
+            self.source[line] = ''
+
     def fix_e303(self, result):
         delete_linenum = int(result['info'].split("(")[1].split(")")[0]) - 2
         delete_linenum = max(1, delete_linenum)
@@ -289,21 +303,26 @@ class FixPEP8(object):
         # are comments.
         cnt = 0
         line = result['line'] - 2
+        modified_lines = []
         while cnt < delete_linenum:
             if line < 0:
                 break
             if not self.source[line].strip():
                 self.source[line] = ''
+                modified_lines.append(line)
                 cnt += 1
             line -= 1
 
+        return modified_lines
+
     def fix_e401(self, result):
         line_index = result['line'] - 1
         target = self.source[line_index]
 
         # Take care of semicolons first
         if ';' in target:
-            self.source[line_index] = _fix_multiple_statements(target)
+            self.source[line_index] = _fix_multiple_statements(target,
+                                                               self.newline)
         else:
             indentation = target.split("import ")[0]
             modules = target.split("import ")[1].split(",")
@@ -321,17 +340,18 @@ class FixPEP8(object):
 
     def fix_e702(self, result):
         target = self.source[result['line'] - 1]
-        self.source[result['line'] - 1] = _fix_multiple_statements(target)
+        self.source[result['line'] - 1] = _fix_multiple_statements(target,
+                self.newline)
 
     def fix_w291(self, result):
         fixed_line = self.source[result['line'] - 1].rstrip()
         self.source[result['line'] - 1] = "%s%s" % (fixed_line, self.newline)
 
     def fix_w292(self, _):
-        # Missing newline gets fixed automatically
-        pass
+        self.source[-1] += self.newline
 
     def fix_w293(self, result):
+        assert not self.source[result['line'] - 1].strip()
         self.source[result['line'] - 1] = self.newline
 
     def fix_w391(self, _):
@@ -340,6 +360,7 @@ class FixPEP8(object):
         found_notblank = False
         blank_count = 0
         for line in source:
+            line = line.rstrip()
             if re.match("^$", line):
                 blank_count += 1
             else:
@@ -351,7 +372,47 @@ class FixPEP8(object):
         self.source = source
 
     def fix_w601(self, result):
-        pass
+        target = self.source[result['line'] - 1]
+        _before = ""
+        _after = ""
+        _symbol = ""
+
+        # Skip complicated cases
+        if target.count('(') > 1 or target.count(')') > 1:
+            return
+
+        _tmp = target.split(".has_key", 1)
+
+        # find dict symbol
+        _target = _tmp[0]
+        for offset, t in enumerate(_target[::-1]):
+            if t == " ":
+                _before = _target[::-1][:offset - 1:-1]
+                break
+            else:
+                _symbol = t + _symbol
+
+        # find arg of has_key
+        _target = _tmp[1]
+        _level = 0
+        _arg = ""
+        for offset, t in enumerate(_target):
+            if t == "(":
+                _level += 1
+            elif t == ")":
+                _level -= 1
+                if _level == 0:
+                    _after += _target[offset + 1:]
+                    break
+            else:
+                _arg += t
+
+        # Maintain precedence
+        if ' ' in _arg.strip():
+            _arg = '(' + _arg.strip() + ')'
+
+        self.source[result['line'] - 1] = \
+            "".join([_before, _arg, " in ", _symbol, _after])
 
     def fix_w602(self, result):
         line_index = result['line'] - 1
@@ -359,19 +420,58 @@ class FixPEP8(object):
 
         if ';' in line:
             # Take care of semicolons first
-            self.source[line_index] = _fix_multiple_statements(line)
+            self.source[line_index] = _fix_multiple_statements(line,
+                                                               self.newline)
             return
-        elif len(line) >= 2 and line[-2] == '\\':
-            # Remove escaped newlines first
+        elif line[-2:] == '\\\n':
+            # Remove escaped LF first
             self.source[line_index] = line[:-2]
             return
-        elif '"""' in line:
-            # FIXME: We currently can't handle
-            #
-            #            raise ValueError, """
-            #                blah
-            #                """
+        elif line[-3:] == '\\\r\n':
+            # Remove escaped CRLF first
+            self.source[line_index] = line[:-3]
             return
+        elif line[-2:] == '\\\r':
+            # Remove escaped CR first
+            # NOTE: Doesn't get executed because pep8 doesn't seem to work with
+            #       CR line endings
+            self.source[line_index] = line[:-2]
+            return
+
+        modified_lines = [line_index]
+
+        double = '"""'
+        single = "'''"
+        if double in line or single in line:
+            # Move full multiline string to current line
+            if double in line and single in line:
+                quotes = (double if line.find(double) < line.find(single)
+                          else single)
+            elif double in line:
+                quotes = double
+            else:
+                quotes = single
+            assert quotes in line
+
+            # Find last line of multiline string
+            end_line_index = line_index
+            if line.count(quotes) == 1:
+                for i in range(line_index + 1, len(self.source)):
+                    end_line_index = i
+                    if quotes in self.source[i]:
+                        break
+
+            # We do not handle anything other than plain multiline strings
+            if ('(' in self.source[end_line_index] or
+                '\\' in self.source[end_line_index]):
+                return []
+
+            for i in range(line_index + 1, end_line_index + 1):
+                line_contents = self.source[i]
+                self.source[line_index] += line_contents
+                self.source[i] = ''
+                modified_lines.append(i)
+            line = self.source[line_index]
 
         sio = StringIO(line)
         is_found_raise = False
@@ -408,15 +508,22 @@ class FixPEP8(object):
                      '(',
                      args[1:-1] if args.startswith('(') else args,
                      ')',
-                     comment, '\n'])
+                     comment, self.newline])
+
+        return modified_lines
 
     def fix_w603(self, result):
         target = self.source[result['line'] - 1]
         self.source[result['line'] - 1] = re.sub('<>', '!=', target)
 
     def fix_w604(self, result):
-        # FIXME: b = ```a``` + ```a```
         target = self.source[result['line'] - 1]
+
+        # We do not support things like
+        #     ``1`` + ``1``
+        if len(re.findall('`+', target)) > 2:
+            return
+
         start = target.find('`')
         end = target[::-1].find('`') * -1
         self.source[result['line'] - 1] = "%srepr(%s)%s" % (
@@ -453,11 +560,11 @@ def _get_indentword(source):
     return indent_word
 
 
-def _fix_multiple_statements(target):
+def _fix_multiple_statements(target, newline):
     non_whitespace_index = len(target) - len(target.lstrip())
     indentation = target[:non_whitespace_index]
     f = [indentation + t.strip() for t in target.split(";") if t.strip()]
-    return '\n'.join(f) + '\n'
+    return newline.join(f) + newline
 
 
 def _analyze_pep8result(result):
@@ -477,6 +584,41 @@ def _get_difftext(old, new, filename):
     return "".join(difftext)
 
 
+def fix_file(filename, opts):
+    tmp_source = read_from_filename(filename)
+    fix = FixPEP8(filename, opts, contents=tmp_source)
+    fixed_source = fix.fix()
+    original_source = copy.copy(fix.original_source)
+    tmp_filename = filename
+    for _ in range(opts.pep8_passes):
+        if fixed_source == tmp_source:
+            break
+        tmp_source = copy.copy(fixed_source)
+        if not pep8:
+            tmp_filename = tempfile.mkstemp()[1]
+            fp = open(tmp_filename, 'w')
+            fp.write(fixed_source)
+            fp.close()
+        fix = FixPEP8(tmp_filename, opts, contents=tmp_source)
+        fixed_source = fix.fix()
+        if not pep8:
+            os.remove(tmp_filename)
+    del tmp_filename
+    del tmp_source
+
+    if opts.diff:
+        new = StringIO("".join(fix.source))
+        new = new.readlines()
+        sys.stdout.write(_get_difftext(original_source, new,
+                                       filename))
+    elif opts.in_place:
+        fp = open(filename, 'w')
+        fp.write(fixed_source)
+        fp.close()
+    else:
+        sys.stdout.write(fixed_source)
+
+
 def main():
     """tool main"""
     parser = OptionParser(version="autopep8: %s" % __version__,
@@ -485,39 +627,23 @@ def main():
                       help='print to verbose result.')
     parser.add_option('-d', '--diff', action='store_true', dest='diff',
                       help='diff print of fixed source.')
-    parser.add_option('-p', '--pep8-passes', default=5, type='int',
+    parser.add_option('-i', '--in-place', action='store_true',
+                      help='make changes to files in place')
+    parser.add_option('-p', '--pep8-passes', default=20, type='int',
                       help='maximum number of additional pep8 passes')
     parser.add_option('--ignore', default='',
                       help='do not fix these errors/warnings (e.g. E4,W)')
     opts, args = parser.parse_args()
     if not len(args):
-        print parser.format_help()
+        print(parser.format_help())
         return 1
-    filename = args[0]
-    original_filename = filename
-    tmp_source = read_from_filename(filename)
-    fix = FixPEP8(filename, opts, contents=tmp_source)
-    fixed_source = fix.fix()
-    original_source = copy.copy(fix.original_source)
-    for _ in range(opts.pep8_passes):
-        if fixed_source == tmp_source:
-            break
-        tmp_source = copy.copy(fixed_source)
-        if not pep8:
-            filename = tempfile.mkstemp()[1]
-            fp = open(filename, 'w')
-            fp.write(fixed_source)
-            fp.close()
-        fix = FixPEP8(filename, opts, contents=tmp_source)
-        fixed_source = fix.fix()
-        if not pep8:
-            os.remove(filename)
-    if opts.diff:
-        new = StringIO("".join(fix.source))
-        new = new.readlines()
-        print _get_difftext(original_source, new, original_filename),
+
+    if opts.in_place:
+        for f in args:
+            fix_file(f, opts)
     else:
-        print fixed_source,
+        fix_file(args[0], opts)
+
 
 if __name__ == '__main__':
     main()
